@@ -65,6 +65,7 @@ struct TempTraversal {
     id: String,
     curve: SphericalLineStringCurve,
     segments: Vec<SegmentOfTraversal>,
+    lrms: Vec<usize>,
 }
 
 impl TempTraversal {
@@ -248,6 +249,7 @@ impl<'fbb> Builder<'fbb> {
             id: traversal_id.to_owned(),
             curve: SphericalLineStringCurve::new(geo::LineString::new(coords), 100.),
             segments: segments.to_vec(),
+            lrms: vec![],
         });
         self.nodes_of_traversal.push(nodes_of_traversal);
 
@@ -279,6 +281,9 @@ impl<'fbb> Builder<'fbb> {
             projected_anchors: Some(self.project_anchors(&anchors, traversal_index)),
             ..Default::default()
         };
+        self.temp_traversal[traversal_index]
+            .lrms
+            .push(self.lrms.len());
         self.lrms
             .push(LinearReferencingMethod::create(&mut self.fbb, &args));
     }
@@ -381,9 +386,18 @@ impl<'fbb> Builder<'fbb> {
 
     /// Return the binary data.
     pub fn build_data(&mut self, properties: Properties) -> &[u8] {
+        use geo_index::rtree::RTreeBuilder;
+        use geo_index::rtree::sort::HilbertSort;
+
         let segments = self.build_segments();
         let traversals = self.build_traversals();
 
+        let mut rtree_builder = RTreeBuilder::<f64>::new(traversals.len() as u32);
+        for traversal in &self.temp_traversal {
+            rtree_builder.add_rect(&traversal.curve.bbox());
+        }
+        let tree = rtree_builder.finish::<HilbertSort>();
+
         let lrs_args = LrsArgs {
             properties: self.build_properties(properties),
             nodes: Some(self.fbb.create_vector(&self.nodes)),
@@ -392,6 +406,7 @@ impl<'fbb> Builder<'fbb> {
             anchors: Some(self.fbb.create_vector(&self.anchors)),
             linear_referencing_methods: Some(self.fbb.create_vector(&self.lrms)),
             geometry_type: GeometryType::Geographic,
+            lrm_spatial_index: Some(self.fbb.create_vector(&tree.into_inner())),
         };
 
         let lrs = Lrs::create(&mut self.fbb, &lrs_args);
@@ -534,14 +549,14 @@ impl<'fbb> Builder<'fbb> {
 }
 
 #[cfg(test)]
-mod tests {
+pub(crate) mod tests {
     use crate::lrm_scale::LrmScaleMeasure;
 
     use super::*;
     use approx::assert_relative_eq;
     use geo::{coord, point};
 
-    fn build_traversal(builder: &mut Builder) -> usize {
+    pub fn build_traversal(builder: &mut Builder) -> usize {
         let s1 = builder.add_segment("s1", &[coord! {x: 0., y: 0.}, coord! {x:1., y: 0.}], 0, 1);
         let s2 = builder.add_segment("s2", &[coord! {x: 1., y: 0.}, coord! {x:2., y: 0.}], 1, 2);
         let sot1 = super::SegmentOfTraversal {
