@@ -10,6 +10,7 @@ use std::cmp::Ordering;
 
 use flatbuffers::{ForwardsUOffset, Vector};
 use geo::orient::Direction;
+use geo_index::rtree::{RTreeIndex, RTreeRef};
 use thiserror::Error;
 
 use crate::curves::{Curve, CurveError};
@@ -17,7 +18,7 @@ use crate::lrm_scale::{
     Anchor, CurvePosition, LrmScale, LrmScaleError, LrmScaleMeasure, ScalePosition,
 };
 use crate::lrs_generated;
-use geo::{LineString, Point, coord, point};
+use geo::{Contains, LineString, Point, coord, point};
 
 /// Used as handle to identify a [`LrmScale`] within a specific [`Lrs`].
 #[derive(Copy, Clone, Debug, Eq, Hash, PartialEq)]
@@ -35,10 +36,8 @@ pub struct NodeHandle(pub usize);
 pub struct Lrm {
     /// The scale of this [`Lrm`].
     pub scale: LrmScale,
-    /// The [`Traversal`] that is the reference of this [`Lrm`].
-    pub reference_traversal: TraversalHandle,
-    /// All the [`Traversal`]s where this [`Lrm`] applies.
-    pub traversals: Vec<TraversalHandle>,
+    /// The [`Traversal`] that where this [`Lrm`] applies.
+    pub traversal: TraversalHandle,
     /// Metadata to describe the Lrm
     pub properties: Properties,
 }
@@ -67,6 +66,8 @@ pub struct Lrs<CurveImpl: Curve> {
     pub nodes: Vec<Node>,
     /// All the [`Segment`] of this Lrs
     pub segments: Vec<Segment>,
+    /// An RTree spatial index of the LRM extents
+    pub rtree_data: Option<Vec<u8>>,
 }
 
 /// A Node is a topological element of the [`Lrs`] that represents a intersection (or an extremity) of an [`Lrm`]
@@ -82,6 +83,25 @@ pub struct Node {
     pub properties: Properties,
 }
 
+impl From<&lrs_generated::Point> for Point {
+    fn from(fb_point: &lrs_generated::Point) -> Self {
+        point! {
+            x: fb_point.x(),
+            y: fb_point.y(),
+        }
+    }
+}
+
+impl From<lrs_generated::Node<'_>> for Node {
+    fn from(fb_node: lrs_generated::Node) -> Self {
+        Self {
+            id: fb_node.id().to_owned(),
+            geometry: fb_node.geometry().map(Point::from),
+            properties: from_fb(fb_node.properties()),
+        }
+    }
+}
+
 /// A segment is a topological element of the [`Lrs`] that represents a piece of the [`Curve`] of an [`Lrm`]
 ///
 /// It has a start and end [`Node`].
@@ -96,6 +116,17 @@ pub struct Segment {
     pub end_node: NodeHandle,
 }
 
+impl From<lrs_generated::Segment<'_>> for Segment {
+    fn from(fb_segment: lrs_generated::Segment) -> Self {
+        Self {
+            id: fb_segment.id().to_owned(),
+            properties: from_fb(fb_segment.properties()),
+            start_node: NodeHandle(fb_segment.start_node_index() as usize),
+            end_node: NodeHandle(fb_segment.end_node_index() as usize),
+        }
+    }
+}
+
 /// The result of a projection onto an [`LrmScale`].
 pub struct LrmProjection {
     /// Contains `measure` ([`LrmScaleMeasure`]) and `lrm` ([`LrmHandle`]).
@@ -195,12 +226,16 @@ impl<CurveImpl: Curve> Lrs<CurveImpl> {
     pub fn from_bytes(buf: &[u8]) -> Result<Self, LrsError> {
         let lrs = lrs_generated::root_as_lrs(buf).map_err(LrsError::InvalidArchive)?;
 
+        let rtree_data = lrs
+            .lrm_spatial_index()
+            .map(|buffer| buffer.bytes().to_vec());
         let mut result = Self {
             lrms: vec![],
             traversals: vec![],
             properties: from_fb(lrs.properties()),
             nodes: vec![],
             segments: vec![],
+            rtree_data,
         };
 
         let source_anchors = lrs
@@ -216,7 +251,7 @@ impl<CurveImpl: Curve> Lrs<CurveImpl> {
                     .get(segment.segment_index() as usize)
                     .geometry()
                     .iter()
-                    .map(|p| (coord! {x: p.x(),y: p.y()}))
+                    .map(|p| coord! {x: p.x(),y: p.y()})
                     .collect();
                 if segment.direction() == lrs_generated::Direction::Decreasing {
                     geom.reverse();
@@ -240,12 +275,12 @@ impl<CurveImpl: Curve> Lrs<CurveImpl> {
             .iter()
             .enumerate()
         {
-            let reference_traversal_idx = raw_lrm.traversal_index() as usize;
+            let traversal_idx = raw_lrm.traversal_index() as usize;
             let curve = &result
                 .traversals
-                .get(reference_traversal_idx)
+                .get(traversal_idx)
                 .ok_or(LrsError::IncompleteArchive(format!(
-                    "traversal {reference_traversal_idx} from lrm {lrm_idx}"
+                    "traversal {traversal_idx} from lrm {lrm_idx}"
                 )))?
                 .curve;
 
@@ -261,18 +296,13 @@ impl<CurveImpl: Curve> Lrs<CurveImpl> {
                         .projected_anchors()
                         .map(|anchors| {
                             let projected_anchor = anchors.get(idx);
-                            let geometry = projected_anchor.geometry().map(|geom| {
-                                point! {
-                                    x: geom.x(),
-                                    y: geom.y(),
-                                }
-                            });
+                            let geometry = projected_anchor.geometry().map(Point::from);
                             (projected_anchor.distance_along_curve(), geometry)
                         })
                         .unwrap_or_else(|| project(&anchor, curve));
 
                     match anchor.name() {
-                        Some(name) => Anchor::new(
+                        Some(name) => Anchor::new_named(
                             name,
                             scale_position,
                             curve_position,
@@ -294,34 +324,23 @@ impl<CurveImpl: Curve> Lrs<CurveImpl> {
                     id: raw_lrm.id().to_owned(),
                     anchors,
                 },
-                reference_traversal: TraversalHandle(reference_traversal_idx),
-                traversals: vec![TraversalHandle(reference_traversal_idx)],
+                traversal: TraversalHandle(traversal_idx),
                 properties: from_fb(raw_lrm.properties()),
             };
 
+            result.traversals[traversal_idx]
+                .lrms
+                .push(LrmHandle(lrm_idx));
+
             result.lrms.push(lrm);
         }
 
         for raw_node in lrs.nodes().unwrap_or_default().iter() {
-            result.nodes.push(Node {
-                id: raw_node.id().to_owned(),
-                geometry: raw_node.geometry().map(|geom| {
-                    point! {
-                        x: geom.x(),
-                        y: geom.y(),
-                    }
-                }),
-                properties: from_fb(raw_node.properties()),
-            });
+            result.nodes.push(Node::from(raw_node));
         }
 
         for raw_segment in lrs.segments().unwrap_or_default().iter() {
-            result.segments.push(Segment {
-                id: raw_segment.id().to_owned(),
-                properties: from_fb(raw_segment.properties()),
-                start_node: NodeHandle(raw_segment.start_node_index() as usize),
-                end_node: NodeHandle(raw_segment.end_node_index() as usize),
-            })
+            result.segments.push(Segment::from(raw_segment))
         }
         Ok(result)
     }
@@ -376,28 +395,22 @@ pub trait LrsBase {
     /// it will be segmentized as a [`LineString`] and might not be as acurate as the underlying representation
     fn get_linestring(&self, traversal: TraversalHandle) -> Result<LineString, LrsError>;
 
-    /// Projects a [`Point`] on all applicable [`Traversal`]s to a given [`Lrm`].
+    /// Projects a [`Point`] on the [`Traversal`]s to a given [`Lrm`].
     /// The [`Point`] must be in the bounding box of the [`Curve`] of the [`Traversal`].
-    /// The result is sorted by `orthogonal_offset`: the nearest [`Lrm`] to the [`Point`] is the first item.
-    fn lookup(&self, point: Point, lrm: LrmHandle) -> Vec<LrmProjection>;
+    fn lookup(&self, point: Point, lrm: LrmHandle) -> Result<LrmProjection, LrsError>;
     /// Projects a [`Point`] on all [`Lrm`] where the [`Point`] is in the bounding box.
     /// The result is sorted by `orthogonal_offset`: the nearest [`Lrm`] to the [`Point`] is the first item.
     fn lookup_lrms(&self, point: Point) -> Vec<LrmProjection>;
-    /// Projects a [`Point`] on all [`Traversal`]s where the [`Point`] is in the bounding box.
-    /// The result is sorted by `orthogonal_offset`: the nearest [`Lrm`] to the [`Point`] is the first item.
-    fn lookup_traversals(&self, point: Point) -> Vec<TraversalProjection>;
+    /// Returns all the traversals whose bounding box include the given point
+    ///
+    /// The function will use the spatial index if it is defined
+    fn traversals_containing(&self, point: Point) -> Vec<TraversalHandle>;
 
     /// Given a [`TraversalPosition`], returns it geographical position ([`Point`]).
     fn locate_traversal(&self, position: TraversalPosition) -> Result<Point, LrsError>;
 
-    /// And [`Lrm`] can be used on many [`Traversal`]s.
-    /// For example, for both directions of a highway.
-    /// This method returns all applicable [`TraversalHandle`]s to that [`Traversal`].
-    fn get_lrm_applicable_traversals(&self, lrm: LrmHandle) -> &[TraversalHandle];
-    /// An [`Lrm`] has a reference [`Traversal`]
-    /// For example, for the centerline of a highway, or a specific track.
-    /// This methods returns the [`TraversalHandle`].
-    fn get_lrm_reference_traversal(&self, lrm: LrmHandle) -> TraversalHandle;
+    /// This methods returns the [`TraversalHandle`] of the [`Lrm`].
+    fn get_lrm_traversal(&self, lrm: LrmHandle) -> TraversalHandle;
 
     /// A [`Traversal`] can be use for multiple [`Lrm`]s.
     /// For example, a highway could have milestones referenced in `miles` AND `kilometers`.
@@ -460,61 +473,56 @@ impl<CurveImpl: Curve> LrsBase for Lrs<CurveImpl> {
             .map(TraversalHandle)
     }
 
-    fn lookup(&self, point: Point, lrm_handle: LrmHandle) -> Vec<LrmProjection> {
+    fn lookup(&self, point: Point, lrm_handle: LrmHandle) -> Result<LrmProjection, LrsError> {
         let lrm = &self.lrms[lrm_handle.0];
-        let mut result: Vec<_> = self
-            .get_lrm_applicable_traversals(lrm_handle)
-            .iter()
-            .flat_map(|t| self.traversals[t.0].curve.project(point))
-            .flat_map(|projection| {
-                let measure = lrm.scale.locate_anchor(projection.distance_along_curve)?;
-                Ok::<LrmProjection, LrsError>(LrmProjection {
-                    measure: LrmMeasure {
-                        lrm: lrm_handle,
-                        measure,
-                    },
-                    orthogonal_offset: projection.offset,
-                })
+        let traversal = &self.traversals[lrm.traversal.0];
+        let projection = traversal.curve.project(point)?;
+        let measure = lrm.scale.locate_anchor(projection.distance_along_curve)?;
+        Ok(LrmProjection {
+            measure: LrmMeasure {
+                lrm: lrm_handle,
+                measure,
+            },
+            orthogonal_offset: projection.offset,
+        })
+    }
+
+    fn traversals_containing(&self, point: Point) -> Vec<TraversalHandle> {
+        let rtree = self
+            .rtree_data
+            .as_ref()
+            .and_then(|buf| RTreeRef::try_new(buf).ok());
+
+        rtree
+            .map(|tree| {
+                // The rectangles inserted in the rtree are already expanded by the margin
+                // That is why the min and max values are the same.
+                tree.search(point.x(), point.y(), point.x(), point.y())
+                    .iter()
+                    .map(|idx| TraversalHandle(*idx as usize))
+                    .collect()
             })
-            .collect();
-        result.sort_by(|a, b| {
-            a.orthogonal_offset
-                .partial_cmp(&b.orthogonal_offset)
-                .unwrap_or(Ordering::Equal)
-        });
-        result
+            .unwrap_or(
+                self.traversals
+                    .iter()
+                    .enumerate()
+                    .filter(|(_idx, traversal)| traversal.curve.bbox().contains(&point))
+                    .map(|(idx, _traversal)| TraversalHandle(idx))
+                    .collect(),
+            )
     }
 
     fn lookup_lrms(&self, point: Point) -> Vec<LrmProjection> {
         let mut result: Vec<_> = self
-            .lrms
+            .traversals_containing(point)
             .iter()
-            .enumerate()
-            .flat_map(|(lrm_idx, _lrm)| self.lookup(point, LrmHandle(lrm_idx)))
+            .flat_map(|traversal_handle| &self.traversals[traversal_handle.0].lrms)
+            .flat_map(|&lrm_handle| self.lookup(point, lrm_handle))
             .collect();
         result.sort_by(|a, b| {
             a.orthogonal_offset
-                .partial_cmp(&b.orthogonal_offset)
-                .unwrap_or(Ordering::Equal)
-        });
-        result
-    }
-
-    fn lookup_traversals(&self, point: Point) -> Vec<TraversalProjection> {
-        let mut result: Vec<_> = self
-            .traversals
-            .iter()
-            .enumerate()
-            .flat_map(|(idx, traversal)| traversal.curve.project(point).map(|proj| (idx, proj)))
-            .map(|(idx, proj)| TraversalProjection {
-                traversal: TraversalHandle(idx),
-                orthogonal_offset: proj.offset,
-                distance_from_start: proj.distance_along_curve,
-            })
-            .collect();
-        result.sort_by(|a, b| {
-            a.orthogonal_offset
-                .partial_cmp(&b.orthogonal_offset)
+                .abs()
+                .partial_cmp(&b.orthogonal_offset.abs())
                 .unwrap_or(Ordering::Equal)
         });
         result
@@ -526,12 +534,8 @@ impl<CurveImpl: Curve> LrsBase for Lrs<CurveImpl> {
             .resolve(position.curve_position)?)
     }
 
-    fn get_lrm_applicable_traversals(&self, lrm: LrmHandle) -> &[TraversalHandle] {
-        &self.lrms[lrm.0].traversals
-    }
-
-    fn get_lrm_reference_traversal(&self, lrm: LrmHandle) -> TraversalHandle {
-        self.lrms[lrm.0].reference_traversal
+    fn get_lrm_traversal(&self, lrm: LrmHandle) -> TraversalHandle {
+        self.lrms[lrm.0].traversal
     }
 
     fn get_traversal_lrms(&self, traversal: TraversalHandle) -> &[LrmHandle] {
@@ -725,16 +729,14 @@ mod tests {
         };
 
         let lrm = Lrm {
-            reference_traversal: TraversalHandle(0),
             scale: crate::lrm_scale::tests::scale(),
-            traversals: vec![TraversalHandle(0)],
+            traversal: TraversalHandle(0),
             properties: properties!("some key" => "some value"),
         };
 
         let mut lrm2 = Lrm {
-            reference_traversal: TraversalHandle(0),
+            traversal: TraversalHandle(1),
             scale: crate::lrm_scale::tests::scale(),
-            traversals: vec![TraversalHandle(0), TraversalHandle(1)],
             properties: properties!(),
         };
         "id2".clone_into(&mut lrm2.scale.id);
@@ -745,6 +747,7 @@ mod tests {
             properties: properties!("source" => "test"),
             nodes: vec![],
             segments: vec![],
+            rtree_data: None,
         }
     }
 
@@ -768,40 +771,34 @@ mod tests {
     #[test]
     fn lookup_single_lrm() {
         let lrs = lrs();
-        let result = lrs.lookup(point! {x: 50., y:0.5}, lrs.get_lrm("id").unwrap());
-        assert_eq!(result.len(), 1);
-        assert_eq!(result[0].orthogonal_offset, 0.5);
-        assert_eq!(result[0].measure.measure.scale_offset, 5.);
+        let result = lrs
+            .lookup(point! {x: 50., y:0.5}, lrs.get_lrm("id").unwrap())
+            .unwrap();
+        assert_eq!(result.orthogonal_offset, 0.5);
+        assert_eq!(result.measure.measure.scale_offset, 5.);
     }
 
     #[test]
     fn lookup_multiple_lrm() {
         let lrs = lrs();
-        let result = lrs.lookup(point! {x: 50., y:0.5}, lrs.get_lrm("id2").unwrap());
-        assert_eq!(result.len(), 2);
-        assert_eq!(result[0].orthogonal_offset, 0.5);
-        assert_eq!(result[0].measure.measure.scale_offset, 5.);
-        assert_eq!(result[0].measure.measure.anchor_name, "a");
-        assert_eq!(result[1].orthogonal_offset, 1.5);
-        assert_eq!(result[1].measure.measure.scale_offset, 5.);
+        let result = lrs
+            .lookup(point! {x: 50., y:0.5}, lrs.get_lrm("id2").unwrap())
+            .unwrap();
+        assert_eq!(result.orthogonal_offset, 1.5);
+        assert_eq!(result.measure.measure.scale_offset, 5.);
+        assert_eq!(result.measure.measure.anchor_name, "a");
+        assert_eq!(result.orthogonal_offset, 1.5);
+        assert_eq!(result.measure.measure.scale_offset, 5.);
     }
 
     #[test]
     fn lookup_lrms() {
         let result = lrs().lookup_lrms(point! {x: 50., y:0.5});
-        assert_eq!(result.len(), 3);
+        assert_eq!(result.len(), 2);
         assert_eq!(result[0].orthogonal_offset, 0.5);
         assert_eq!(result[0].measure.measure.scale_offset, 5.);
-        assert_eq!(result[1].orthogonal_offset, 0.5);
+        assert_eq!(result[1].orthogonal_offset, 1.5);
         assert_eq!(result[1].measure.measure.scale_offset, 5.);
-        assert_eq!(result[2].orthogonal_offset, 1.5);
-        assert_eq!(result[2].measure.measure.scale_offset, 5.);
-    }
-
-    #[test]
-    fn lookup_traversals() {
-        let result = lrs().lookup_traversals(point! {x: 50., y:0.5});
-        assert_eq!(result.len(), 2);
     }
 
     #[test]
@@ -824,21 +821,11 @@ mod tests {
     }
 
     #[test]
-    fn get_lrm_applicable_traversals() {
-        let lrs = lrs();
-        let result = lrs.get_lrm_applicable_traversals(LrmHandle(0));
-        assert_eq!(result.len(), 1);
-
-        let result = lrs.get_lrm_applicable_traversals(LrmHandle(1));
-        assert_eq!(result.len(), 2);
-    }
-
-    #[test]
-    fn get_lrm_reference_traversal() {
-        let result = lrs().get_lrm_reference_traversal(LrmHandle(0));
-        assert_eq!(TraversalHandle(0), result);
-        let result = lrs().get_lrm_reference_traversal(LrmHandle(1));
+    fn get_lrm_traversal() {
+        let result = lrs().get_lrm_traversal(LrmHandle(0));
         assert_eq!(TraversalHandle(0), result);
+        let result = lrs().get_lrm_traversal(LrmHandle(1));
+        assert_eq!(TraversalHandle(1), result);
     }
 
     #[test]
